definition module Prelude

from StdOverloaded import 
     class ==(..),		class <(..),    class +(..),    	class -(..),		class *(..),    class /(..)
    ,class abs(..),		class zero(..), class one(..),  	class exp(..)
    ,class sin(..),		class cos(..),  class asin(..), 	class acos(..),		class atan(..)
    ,class sinh(..),	class cosh(..), class asinh(..),	class acosh(..),	class atanh(..)
    ,class fromInt(..),	class sign(..), class toReal(..),	class fromReal(..),	class ln(..)
    ,class toChar(..),	class fromChar(..), 				class fromString(..), class toString(..)
    ,class +++(..),		class ~(..)
    
from StdBool        import  not,||,&& ,instance == Bool

import  intInstances
import realInstances
import charInstances

from PreludeList    import  foldr,map,repeat,concat,iterate,takeWhile
from Ratio          import  :: Rational, :: Ratio(..)
from PreludeText    import  class Show(..), :: ShowS(..)

from PreludeLayer   import  from2,from_to,from_then_to

// =====================================================================================================

// -------- EQ class ----------- 

class Eq a | == a
where
 (/=) infix  4 :: !a    !a  ->  Bool | Eq a
 (/=) x y :== not (x == y)

// -------- ORD class ----------- 

class Ord a | Eq, < a
where
    compare :: !a !a -> Ordering     | Ord a
    compare x y :== if (x==y) EQ (if (x<y) LT GT)
                                
    (>) infix  4 :: !a !a   ->  Bool | Ord a
    (>) x y  :== y < x 
    
    (<=) infix  4 :: !a !a  ->  Bool | Ord a
    (<=) x y :== not (y<x)
    
    (>=) infix  4 :: !a !a  ->  Bool | Ord a
    (>=) x y :== not (x<y) 
    
    min::!a !a ->   a | Ord a
    min x y  :== if (x<y) x y
    
    max::!a !a ->   a | Ord a
    max x y  :== if (x<y) y x

// -------- ENUM class -----------    

class Enum a   
where
    toEnum          :: Int -> a
    fromEnum        :: a -> Int
    enumFromThen    :: a a -> [a]                       // [n,n`..]
    
    succ    :: !a -> a | Enum a
    succ  x :== toEnum ((fromEnum x) + 1)
    
    pred    :: !a -> a | Enum a
    pred  x :== toEnum ((fromEnum x) - 1)

    enumFrom :: !a -> [a]  | Enum a                     // [n..]
    enumFrom x :== map toEnum (from2 (fromEnum x)) 
    
    enumFromTo :: !a !a -> [a] | Enum a                 // [n..m]
    enumFromTo x y :== map toEnum (from_to (fromEnum x) (fromEnum y))
    
    enumFromThenTo :: !a !a !a -> [a] | Enum a          // [n,n`..m]
    enumFromThenTo x y z :== map toEnum (from_then_to (fromEnum x) (fromEnum y) (fromEnum z))

// -------- BOUNDED class ----------- 

class Bounded a 
where
    minBound :: a
    maxBound :: a

// -------- NUM class ----------- 

class Num a | Eq,+,-,*,abs,fromInt,zero,one,Show a
where
    signum      :: a -> a

    negate      ::  a -> a  | Num a
    negate x    :== zero - x
/*  fromInteger     :: Integer -> a
                not supported yet */

// -------- REAL class -----------

class Hs__Real a | Num, Ord a // shoulc be Real : name-space bug clean 2.0 
where
    toRational :: a -> Rational

// -------- INTEGRAL class -----------

class  Integral a | Hs__Real,Enum a
where
    quotRem         :: a a -> (a,a)
    toInteger       :: a -> Int
    
    (quot)  infixl 7    :: a a -> a | Integral a
    (quot) n d          :==  fst (quotRem n d)
    
    (rem) infixl 7      :: a a -> a | Integral a
    (rem) n d           :==  snd (quotRem n d)
    
    (div) infixl 7      :: a a -> a | Integral a
    (div) n d           :==  fst (divMod n d)
    
    (mod) infixl 7      :: a a -> a | Integral a
    (mod) n d           :==  snd (divMod n d)
    
    divMod              :: a a -> (a,a) | Integral a
    divMod n d  :==  
        let (q,r) = quotRem n d
        in
            if ( signum r == ( negate (signum d) )) (q-one, r+d) (q,r)

// -------- FRACTIONAL class -----------    

class Fractional a  | Num,/ a
where
    fromRational     :: Rational -> a
    recip   :: a -> a   | Fractional a
    recip x :==  one / x
   
// -------- FLOATING class -----------    
half :: a | one,/,+ a 
//half      = one / (one+one)
minusone :: a | zero,-,one a
//minusone  = zero - one
two :: a | one, + a
//two       = one + one

class Floating a | Fractional,exp,sin,cos,asin,acos,atan,sinh,cosh,asinh,acosh,atanh a
where
    pi              :: a
    log             :: a -> a 
    
    (**) infixr 8   :: a a -> a     | Floating a
    (**) x y        :== exp (log x * y)
    
    tan             :: a -> a       | Floating a
    tan  x          :== sin x / cos x
    
    logBase         :: a a -> a     | Floating a
    logBase x y     :== log y / log x
    
    sqrt            :: !a -> a      | Floating a 
    sqrt x          :==  x ** half
        
    tanh            :: a -> a       | Floating a
    tanh x          :==  sinh x / cosh x
  
// -------- REALFRAC class -----------    

class RealFrac a | Hs__Real,Fractional a
where
    properFraction :: a -> (b,a)    | Integral b
    
    truncate    :: a -> b       | RealFrac a & Integral b
    truncate x  :== fst (properFraction x)
    
    round  :: a -> b        | RealFrac a & Integral b
    round x :==
        let
            (n,r) = properFraction x
            m = if (r < zero) (n - one) (n + one)
        in 
            case signum (abs r - half) of
                minusone        -> n
                zero            -> if (even n) n m
                one             -> m

    ceiling :: a -> b       | RealFrac a & Integral b
    ceiling x :== let (n,r) = properFraction x in
                        if (r > zero) (n + one) n

    floor   :: a -> b | RealFrac a & Integral b
    floor x :== let (n,r) = properFraction x in 
                if (r < zero) (n - one) n
                

// -------- REALFLOAT class -----------    
    
class RealFloat a | RealFrac,Floating a
where
    floatRadix      :: a -> Int
    floatDigits     :: a -> Int
    floatRange      :: a -> (Int,Int)
    decodeFloat     :: a -> (Int,Int)
    encodeFloat     :: Int Int -> a
    
    isNaN           :: a -> Bool
    isInfinite      :: a -> Bool
    isDenormalized  :: a -> Bool
    isNegativeZero  :: a -> Bool
    isIEEE          :: a -> Bool
    
    exponent    :: a -> Int | RealFloat a
    exponent x :==
        let (m,n) = decodeFloat x in
        if (m == 0) 0 (n + floatDigits x)
                        
    significand :: a -> a       | RealFloat a
    significand x :==
        let (m,_) = decodeFloat x in
        encodeFloat m (negate floatDigits x)
                        
    scaleFloat :: Int a -> a    | RealFloat a
    scaleFloat k x :==
        let (m,n) = decodeFloat x in
        encodeFloat m (n+k)
                        
    atan2       :: a a -> a | RealFloat a
    atan2 y x   :== atan2` y x

atan2` :: a a -> a | RealFloat a 
/*
atan2` :: a a -> a | RealFloat a 
atan2` y x   
    | (x > zero)                = atan (y/x)
    | x == zero && y > zero     =  pi / two
    | x <  zero && y > zero     =  pi + atan (y/x) 
    |       (x <= zero && y < zero)
        ||  (x <  zero && isNegativeZero y)
        ||  (isNegativeZero x && isNegativeZero y)
                                = negate (atan2 (negate y) x)
    | y == zero && ( x < zero || isNegativeZero x)
                                =  pi    // must be after the previous test on zero y
    | x == zero  && y== zero    =  y     // must be after the other double zero tests
                                =  x + y // x or y is a NaN, return a NaN (via +)
*/

// ----------- Numeric functions ---------------

subtract        :: a a -> a     | Num a 
even            :: !a -> Bool   | Integral a 
odd             :: !a -> Bool   | Integral a 
gcd             :: !a !a -> a   | Integral a
lcm             :: !a !a -> a   | Integral a
(^) infixr 8    :: !a !b -> a   | Num a & Integral b
(^^) infixr 8   :: !a !b -> a   | Fractional a & Integral b
fromIntegral    :: !a -> b      | Integral a & Num b
realToFrac      :: !a -> b      | Hs__Real a & Fractional b

class  Functor f
where
    fmap :: (a -> b) (f a) -> f b

class  Monad m
where
    (>>=) infixl 1  :: (m a) (a -> m b) -> m b
    return          :: a -> m a
    fail            :: [Char] -> m a

    (>>) infixl 1   :: (m a) (m b)-> m b  | Monad m
    (>>) m k        :==  m >>= \_ -> k

sequence    :: [m a] -> m [a]       | Monad m 
sequence_   :: [m a] -> m Trivial   | Monad m 

// The xxxM functions take list arguments, but lift the function or list element to a monad type

mapM    :: (a -> m b) [a] -> m [b]          | Monad m
mapM_   :: (a -> m b) [a] -> m Trivial      | Monad m
(=<<) infixr 1 :: (a -> m b) (m a) -> m b   | Monad m

id      :: !a -> a
const   :: !a b -> a

//(.) infixr 9 :: (b -> c) (a -> b) -> (a -> c) : automatically parsed as "o"

flip :: (a b -> c) b a -> c
(seq) infixr 0  :: !a  b -> b
($) infixr 0    :: (a -> b) a -> b
($!) infixr 0   :: (a -> b) !a -> b

// ----- Trivial data type ---------- 

:: Trivial = Trivial 

instance ==      Trivial
instance <       Trivial
instance Enum    Trivial
instance Bounded Trivial
/* See in PreludeGeneric:
instance Show    Trivial
instance Read    Trivial
*/
// --------- BOOL type ------------

//:: Bool  =  False | True
//instance Eq    Bool
instance <       Bool
instance Enum    Bool
instance Bounded Bool
/* See in PreludeGeneric:
instance Show Bool
instance Read Bool
*/

/* 
not                 :: !Bool        ->  Bool
(||)    infixr 2    :: !Bool Bool   ->  Bool
(&&)    infixr 3    :: !Bool Bool   ->  Bool
otherwise :: Bool
*/

// -------------- CHAR type -----------------------

//instance Eq    Char
//instance Ord   Char
instance Enum    Char
instance Bounded Char  

// -------------- MAYBE type -----------------------

:: Maybe a  =   Just a
            |   Nothing

instance ==    (Maybe a) | ==   a
instance <     (Maybe a) | <    a
/* See in PreludeGeneric:
instance Show  (Maybe a) | Show a 
instance Read  (Maybe a) | Read a 
*/

maybe :: b (a -> b) (Maybe a) -> b

instance Functor Maybe
instance Monad Maybe

// -------------- EITHER type -----------------------

:: Either a b   =  Left a 
                | Right b
/* See in PreludeGeneric:
instance ==    (Either a b)  | == a & == b
instance <     (Either a b)  | < a & < b
instance Show  (Either a b)  | Show a & Show b
instance Read  (Either a b)  | Read a & Read b
*/

either :: (a -> c) (b -> c) (Either a b) -> c

/*
// -------------- IO type -----------------------
... can be found in PreludeIO

*/

// ------------- ORDERING type ---------------------
:: Ordering = LT | EQ | GT

instance ==      Ordering
instance <       Ordering
instance Enum    Ordering
instance Bounded Ordering
/* See in PreludeGeneric:
instance Show Ordering
instance Read Ordering
*/
// ---------------- INSTANCES FOR BASIC CLASSES -----------------------------

// --- Int ---
// instance Eq      Int
// instance Ord     Int
instance Num        Int  
instance Hs__Real   Int
instance Enum       Int
instance Integral   Int
instance Bounded    Int

// --- Real (Float)  ---
// instance Eq      Int
// instance Ord     Int
instance Num        Real
instance Hs__Real   Real
instance Fractional Real
instance Floating   Real
instance RealFrac   Real
//!!! instance RealFloat Real

numericEnumFrom         :: a -> [a]     | Fractional a
numericEnumFromThen     :: a a -> [a]   | Fractional a
numericEnumFromTo       :: a a -> [a]   | Fractional, Ord a
numericEnumFromThenTo   :: a a a -> [a] | Fractional, Ord a

numericSucc :: a -> a | +, one a
numericPred :: a -> a | -, one a

instance Enum       Real
// -------- LIST -----------------------
instance == [a] | Eq a
instance <  [a] | Ord a
instance Functor []
instance Monad   []
// --------------------------------------
/* Tuples
data  (a,b)   =  (a,b)
data  (a,b,c) =  (a,b,c)
*/
/* See in PreludeGeneric:
instance Eq  (a,b)
instance Ord (a,b)
instance Eq  (a,b,c)
instance Ord (a,b,c)
*/
instance Bounded (a,b)   | Bounded a & Bounded b
instance Bounded (a,b,c) | Bounded a & Bounded b & Bounded c

// fst      :: !(!.a,.b) -> .a
fst tuple :== t1 where (t1, _) = tuple
// snd      :: !(.a,!.b) -> .b
snd tuple :== t2 where (_, t2) = tuple

curry   :: ((a, b) -> c) a b -> c
uncurry :: (a b -> c) (a, b) -> c


// until p f  yields the result of applying f until p holds.
until :: !(a -> Bool) (a -> a) a -> a

// asTypeOf is a type-restricted version of const.  It is usually used
// as an infix operator, and its typing forces its first argument
// (which is usually overloaded) to have the same type as the second.
asTypeOf  :: !a a -> a

//error stops execution and displays an error message
error :: [Char] -> a

// It is expected that compilers will recognize this and insert error
// messages that are more appropriate to the context in which undefined 
// appears. 
undefined        :: a

// ====================== END OF PRELUDE ==================================================
